/*      Get and Resume Elite EDition source code


Get and Resume Elite EDition (GREED)
Copyright (C) 1999  Anoakie Turner

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

For more information on the GPL, please go to:
http://www.gnu.org/copyleft/gpl.html


        Contact:  Anoakie Turner
                  Anoakie.Turner@asu.edu

                  13240 N. 94th Pl.
                  Scottsdale, AZ 85260
*/

#include "main.h"


void ReadSock (int sockfd, char* buffer, char* stopstr)
/*************************************
**	ReadSock(int, char*, char*)
**
** PRE:	 Assigned(sockfd, buffer, stopstr)
** POST: reads data from socket(sockfd) until either the
**	 stopstr is hit or sockfd reaches the EOF
*************************************/
{	int len = 1;
	int result = 0;

	do
	{	result = read(sockfd, buffer + len - 1,
				sizeof(buffer) + len - 1);
		len += result;
		buffer[len] = '\0';
	} while (result > 0 && strstr(buffer, stopstr) == NULL);
if (OUTPUT_LEVEL > 8)
	printf("%s\n\n", buffer);
}


int ReadPASVPort(char* buffer)
/*************************************
**	ReadPASVPort(char*)
**
** PRE:  Assigned(buffer)
** POST: Returns the first specified valid passive port in (buffer)
*************************************/
{	char *rbuffer = strstr(buffer, "Passive Mode (");
	char lhost[6][4];
	int i = 0, j = 0, k = 0;


	/* If Passive Mode was not found in the buffer,
		we have to assume the ftp server doesn't use
		PASV mode, therefore it fails 
		NOTE: Server could aso be busy! */
	if (rbuffer == NULL)
	{	if (OUTPUT_LEVEL > 0)
			printf("PASV mode failed");
		return(0);
	} else
	{	for (i = 14; (unsigned)i < strlen(rbuffer) && rbuffer[i] != ')'; i++)
			if (rbuffer[i] == ',')
			{	lhost[j][k] = '\0';
				j++;
				k = 0;
			} else
		lhost[j][k++] = rbuffer[i];

		return (atoi(lhost[4])*256 + atoi(lhost[5]));
	}
}


bool ReadFTPListing (URLp URL)
/*************************************
**	bool ReadFTPListing
**
** PRE:  URL has been Parse()'d and Connect()'ed
** POST: Sets URL->sockfd up for reading and determines the files size.
*************************************/
{	int i = 0, j = 0, k = 0, listfd = 0, datafd = 0;
	unsigned int tempport = 0;
	int whitespace = 0;
	char *buffer;
	struct sockaddr_in datastream, liststream;
	URL->size = malloc(15);
	NOSIZE = 0;
	buffer = malloc(6400);

	/* Reads from the FTP socket until 200 (ok) is reached */
	ReadSock(URL->sockfd, buffer, "\n200");

	listfd = socket(AF_INET, SOCK_STREAM, 0);
	tempport = ReadPASVPort(buffer);
	
	/* If we can't find a valid port to connect to, or can't open
		a socket to read the file listing, then an error occurs */
	if (tempport == 0 || listfd == -1)
	{	/* Server returns 530 if busy */
		if (strstr(buffer, "\n530") != NULL || strstr(buffer, "\n421") != NULL)
		{	URL->retry = 1;
			if (OUTPUT_LEVEL > 0)
				printf(" because server is busy");
		}
		else
			URL->retry = 0;
		write(URL->sockfd, "\r\nQUIT\r\n", 8);
		close(URL->sockfd);
		close(listfd);
		if (OUTPUT_LEVEL > 0)
			printf("!\n");
	
		if (OUTPUT_LEVEL > 0 && listfd == -1)
			printf("Error opening socket to FTP server for LIST!\n\n");
	
		return(0);
	}

	liststream.sin_family = AF_INET;
	liststream.sin_port = htons(tempport);
	memcpy(&liststream.sin_addr, URL->host->h_addr, URL->host->h_length);


	/* Attempts to connect to socket for list of files */
	if (connect(listfd, (struct sockaddr *)&liststream, sizeof(liststream)) == -1)
	{	if (OUTPUT_LEVEL > 0)
			printf("Error connecting to FTP socket for LIST!\n");
		write(URL->sockfd, "\r\nQUIT\r\n", 8);
		close(listfd);
		return(0);
	}

	/* Reads the socket for the file listing */
	ReadSock(listfd, buffer, "\r\n\r\n\0");
	close(listfd);

	if (buffer[0] != '-')
		buffer += strchr(buffer, '-') - buffer;

	/* queues the variable 'j' to the start of the filesize */
	/* Sample: "-rwx-wx--x 1 root wheel 1024 /etc/passwd" */
	while ((unsigned)j < strlen(buffer) && whitespace < 4)
	{	if (buffer[j] == ' ' && buffer[j + 1] != ' ')
			whitespace ++;
		j++;
	}	


	k = j;
	/* Calculates the size of the string being read in */
	for (i = 0; (unsigned)j < strlen(buffer); j++)
		if (buffer[j] == ' ' || (buffer[j] > ('0' - 1) && buffer[j] < ('9' + 1)))
			i++;
		else
			j = strlen(buffer);	

	/* Allocates space for the string */
	URL->total = malloc(i + 1);

	j = k;
	k = 0;
	/* reads the first set of numbers it finds into the string URL->total */
	for (i = 0; (unsigned)j < strlen(buffer); j++)
		if (buffer[j] == ' ' || (buffer[j] > ('0' - 1) && buffer[j] < ('9' + 1)))
		{	URL->total[i++] = buffer[j];
			if (buffer[j] != ' ')
				k = 1;
		}
		else
			j = strlen(buffer);


	URL->total[j - 4] = '\0';
	URL->ltotal = atol(URL->total);

	if (OUTPUT_LEVEL > 1 && !isdigit(URL->total[0]))
		printf("File size is unknown!\n");
	else if (OUTPUT_LEVEL > 1)
		printf("File size: %ld bytes\n", URL->ltotal);

	
	write(URL->sockfd, "PASV\r\n", 6);
	usleep(FTP_WAIT_TIME);
	write(URL->sockfd, "REST 0\r\n", 8);
	usleep(FTP_WAIT_TIME);
	write(URL->sockfd, "NOOP\r\n", 6);

	/* Note, it looks for a 200, which NOOP returns */
	ReadSock(URL->sockfd, buffer, "\n200");

	/* If a 350 is returned, then you cannot resume the file! */
	if (strstr(buffer, "\n350") == NULL)
		URL->retry = 0;

	tempport = ReadPASVPort(buffer);
	if (!tempport)
	{	URL->retry = 1;
		printf(" for file retrieval!\n");
		write(URL->sockfd, "\r\nQUIT\r\n", 8);
		close(URL->sockfd);
		return (0);
	}

	/* Opens up the file for appending, seeks to the end of the file and 
		determines teh files size from its current location */

	if (strchr(URL->filename, ' ') != NULL)
		for (i = 0; i < strlen(URL->filename); i++)     
			if (URL->filename[i] == ' ')
				URL->filename[i] = '_';

	URL->fp = fopen (URL->filename, "a+b");
	fseek (URL->fp,0,SEEK_END);
	URL->lsize = ftell(URL->fp);

	/* If the size is of the file (according to the FTP server)
		is 0 bytes, or it doesn't return the file size,
		the file is set to 2 gigs (doesn't effect transfer)
		and the NOSIZE flag is set */
	if (k == 0)
	{	URL->ltotal = 2000000000;
		NOSIZE = 1;
	}


	/* If the size of the file is greater than or equal to the size
		the ftp server reports, then the file is already finished!
		If URL->fp is NULL, then an error has occured when opening
		the file */
	if ((URL->lsize >= URL->ltotal && URL->ltotal != 0) || URL->fp == NULL)
	{	if (OUTPUT_LEVEL > 0)
			printf("Error... It seems this file has already been downloaded OR it's "
				"currently in use!\n");
		write(URL->sockfd, "\r\nQUIT\r\n", 8);
		close(datafd);
		URL->retry = 0;
		return(0);
	}

	/* If the size of the file is smaller then the set ROLLBACK size,
		then the file size is reset to 0, otherwise it rolls
		the file size back ROLLBACK bytes */
	if (URL->lsize < ROLLBACK || !URL->retry)
		URL->lsize = 0;
	else
		URL->lsize -= ROLLBACK;

	sprintf(URL->size, "%ld", URL->lsize);
	if (OUTPUT_LEVEL > 1)
		printf("Resuming from %s bytes.\n\n", URL->size);

	/* Tells the FTP server that you want to resume a file at
		URL->size bytes */
	write(URL->sockfd, "TYPE I\r\n", 8);
	usleep(FTP_WAIT_TIME);
	write(URL->sockfd, "REST ", 5);
	write(URL->sockfd, URL->size, strlen(URL->size));
	write(URL->sockfd, "\r\n", 2);
	usleep(FTP_WAIT_TIME);
	write(URL->sockfd, "RETR ", 5);
	write(URL->sockfd, URL->file, strlen(URL->file));
	write(URL->sockfd, "\r\n", 2);

	/* Sets the dldots equal to the amount that is already downloaded */
	dldots = URL->lsize / ROLLBACK;

	datafd = socket(AF_INET, SOCK_STREAM, 0);
	if (datafd == -1)
	{	if (OUTPUT_LEVEL > 0)
			printf("Error opening socket to FTP server for RETR!\n");
		write(URL->sockfd, "\r\nABOR\r\nQUIT\r\n", 14);
		close(datafd);
		close(URL->sockfd);
		URL->retry = 1;
		return(0);
	}

	datastream.sin_family = AF_INET;
	datastream.sin_port = htons(tempport);
	memcpy(&datastream.sin_addr, URL->host->h_addr, URL->host->h_length);

	if (connect(datafd, (struct sockaddr *)&datastream, sizeof(datastream)) == -1)
	{	if (OUTPUT_LEVEL > 1)
			printf("Error connecting to FTP socket for RETR!\n");
		write(URL->sockfd, "\r\nABOR\r\nQUIT\r\n", 14);
		close(datafd);
		URL->retry = 1;
		return(0);
	}

	ReadSock(URL->sockfd, buffer, "\n350");

	write(URL->sockfd, "\r\nQUIT\r\n", 8);

	listfd = URL->sockfd;
	URL->sockfd = datafd;

	close (listfd);

	URL->lleft = URL->ltotal - URL->lsize;

	if (OUTPUT_LEVEL > 1 && STDOUT == 0)
		printf("[.");

	/* Shrinks the file size for the rollback */
	ftruncate(fileno(URL->fp), URL->lsize);

	return (1);
}
