/*      Get and Resume Elite EDition source code


Get and Resume Elite EDition (GREED)
Copyright (C) 1999  Anoakie Turner

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

For more information on the GPL, please go to:
http://www.gnu.org/copyleft/gpl.html


        Contact:  Anoakie Turner
                  Anoakie.Turner@asu.edu

                  13240 N. 94th Pl.
                  Scottsdale, AZ 85260
*/


/*
**  To compile, use gcc -O2 -Wall -o greed greed.c
**
**
**  To compile under any non linux OS, use:
**  gcc -O2 -Wall -lsocket -lxnet -o greed greed.c
*/


#include "main.h"


int CHECKIN;
int STDIN;
int TESTFILES;
int FORK;
int BACKGROUND;
int IMPORT_GR;
char* GR_FILE;

typedef struct NODE
{	struct URLstruct URL;
	struct NODE *next;
} NODE;
NODE *list;
NODE *curr;


void Command (int argc, char *argv[]);
void Stdin ();
void File ();
void GRFile ();
void StartDownload ();


int main (int argc, char *argv[])
/*************************************
**	int main (int, char*)
**
** main() function.
*************************************/
{	int i = 1;
	list = malloc(sizeof(NODE));
	curr = list;
	curr->next = NULL;
	signal(SIGPIPE, SIG_IGN);

	/* Sets default global variables */
	CHECKIN = 1;
	STDIN = 0;
	FTP_WAIT_TIME = 0;
	OUTPUT_LEVEL = 5;
	STDOUT = 0;
	ROLLBACK = 4096;
	TIMEOUT_SEC = 300;
	MINSPEED = 0;
	REFERRER = 1;
	TESTFILES = 0;
	FORK = 0;
	BACKGROUND = 0;
	IMPORT_GR = 0;
	WAIT_RETRY = 10;

	/* Parses the arguments in argv for switches */
	for (i = 1; i < argc; i ++)
		Switches(i, argv);

	/* Reads in configuration file and resets some globals */
	Proxy();

	/* Prints out Version info */
	if (OUTPUT_LEVEL > 0)
		printf("%s  [Get and Resume Elite EDition] - By "
			"Anoakie Turner\r\n\n", CURRENT_VERSION);

	/* If the background switch was switched, it forks into the background
		and exits the parent */
	if (BACKGROUND)
	{	OUTPUT_LEVEL = 0;
		TESTFILES = 0;
		if (fork() != 0)	/* If I am the parent */
		{	printf("Forking into background mode!\n");
			exit(0);
		}
	}

	/* Parses the URLS in argv */
	if (argc > 1)
		Command (argc, argv);

	/* Parses the URLS read from greed.in */
	if (CHECKIN)
		File ();

	/* Parses the URLS read from a standard GetRight file */
	if (IMPORT_GR)
		GRFile ();

	StartDownload ();

	/* Parses the URLS read in from the standard input */
	if (STDIN)
		Stdin ();

	/* If everything goes well, then exit with EXIT_SUCCESS */
	return(EXIT_SUCCESS);
}


void Command (int argc, char *argv[])
/*************************************
**	void Command (URLp, int, char*)
**
** Pre:  argc == the number of elements in argv.
** Post: Attempts to retrieve the URLs in argv.
*************************************/
{	int i;
	for (i = 1; i < argc; i ++) 
		if (argv[i][0] != '-')
		{	curr->next = malloc(sizeof(NODE));
			curr = curr->next;
			curr->URL.name = argv[i];
			curr->next = NULL;
		}
}


void Stdin ()
/*************************************
**	void Stdin (URLp)
**
** Post: Attempts to retrieve the URLs from STDIN.
** NOTE: currently doesn't support FORK!
*************************************/
{	char *buffer = malloc(1024); 
	struct URLstruct URL;

	printf("Waiting for input...\n");
	while(scanf("%s", buffer) != EOF)
	{	URL.name = buffer;
		if (Parse(&URL))
			DownloadLoop(&URL);
		else
		{	if(OUTPUT_LEVEL > 0)
				printf("Unable to parse %s", buffer);
		}
	}
	free(buffer);
}


void File ()
/*************************************
**	void File (URLp)
**
** Post: Attempts to retrieve the URLs from greed.in.
** NOTE: currently doesn't support FORK!
*************************************/
{	FILE *GD = fopen("greed.in", "r");
	int i = 0;
	char *buffer = malloc(1024);

	if (GD != NULL)
	{	while(!feof(GD))
		{	fread(buffer + i, 1, 1, GD);

		/* If we hit a space or return in the buffer then
		   we have a supposed URL! */
			if(buffer[i] == ' ' || buffer[i] == '\n')
			{	buffer[i] = '\0';
				curr->next = malloc(sizeof(NODE));
				strcpy(curr->URL.name, buffer);
				curr = curr->next;
				curr->next = NULL;

				i = 0;
			}
			else
			{	i++;
			}
		}
		unlink("greed.in");

		fclose(GD);
	}
}


void GRFile ()
/*************************************
**	void GRFile ()
**
** Post: Attempts to retrieve the URLs from a standard .GRX file.
** NOTE: currently doesn't support FORK!
*************************************/
{	FILE *GD = fopen(GR_FILE, "r");
	int i = 0;
	char *buffer = malloc(1024);

	if (GD != NULL)
	{	while(!feof(GD))
		{	fread(buffer + i, 1, 1, GD);

		/* If we hit a '\n' or return in the buffer and it doesn't
		   start with a '/', then we have a supposed URL! */
			if(buffer[0] != '/' && (buffer[i] == '\r' || buffer[i] == '\n'))
			{	buffer[i+1] = '\0';
				if (strstr(buffer, "URL: ") != NULL &&
					strstr(buffer, "PageURL: ") == NULL)
				{	buffer[i] = '\0';

					curr->next = malloc(sizeof(NODE));
					strcpy(curr->URL.name, buffer + 5);
					curr = curr->next;
					curr->next = NULL;

					printf("**** %s ****",curr->URL.name);
				}
				else if (strstr(buffer, "PageURL: ") != NULL)
				{	buffer[i] = '\0';
					strcpy(REF, buffer + 15);
				}
				else if (buffer [0] == '\r' || buffer[0] == '\n')
					i = 0;
				i = 0;
			}
			else
			{	i++;
				if (buffer[i - 1] == '\r' || buffer[i - 1] == '\n')
					i = 0;
			}
		}
		fclose(GD);
	}
}


void StartDownload ()
{	int j;
	int status = 0;
	pid_t childpid = 0;
	curr = list->next;

	if (FORK)
	{	for (j = 1; j < FORK + 1 && !status && curr->next != NULL; j++)
		{	childpid = fork();
			if (childpid == 0)
			{	status = 1;
				curr = curr->next;
			}
		}
	}


	while (curr != NULL)
	{	if(Parse(&curr->URL))
			DownloadLoop(&curr->URL);
		else if(OUTPUT_LEVEL > 0)
			printf("Unable to parse %s", curr->URL.name);
		for (j = 0; j < FORK && curr->next != NULL; j++)
			curr = curr->next;
		curr = curr->next;
	}

	/* If it's the child, then it exits, else it waits until the other
		children have finished what they are doing */
	if (!childpid)
		exit(EXIT_SUCCESS);
	if (childpid == wait(&status) && OUTPUT_LEVEL > 0)
		printf("Waiting for children processes to finish downloading\n");
	while(childpid == wait(&status) && wait(&status) != -1)
		if((childpid == -1) && (errno != EINTR))
			break;
}


void DownloadLoop (URLp URL)
/*************************************
**	void DownloadLoop (URLp)
**
** Pre:  URL has been Parse()'d
** Post: Connects to server and downloads the file
*************************************/
{	char COMMAND[128];
	int ARCHIVES;

	while(URL->retry == 1)
	{	URL->done = 0;
		ARCHIVES = 0;

		if (Connect(URL))
		if ((URL->protocol == FTP && ReadFTPListing(URL)) ||
			(URL->protocol == HTTP && ReadHTTPHeader(URL)))
		{	start = time((time_t *)0);
			end = time((time_t *)0);
			FD_ZERO(&testfds);
			FD_SET(URL->sockfd, &testfds);
			while (!URL->done)
				Download(URL);
		}

		if (OUTPUT_LEVEL > 0)
			printf("\n");
		close (URL->sockfd);

		if (URL->retry == 1)
		{	if (OUTPUT_LEVEL > 1)
				printf("Retrying in %d seconds...\n", WAIT_RETRY);
			sleep(WAIT_RETRY);
		} else if (TESTFILES)
		{	if (strstr(strrchr(URL->filename, '.'), "gz") || strstr(strrchr(URL->filename, '.'), "tgz"))
			{	strcpy(COMMAND, "gzip -t ");
				ARCHIVES = 1;
			} else if (strstr(strrchr(URL->filename, '.'), "tar"))
			{	strcpy(COMMAND, "tar -t ");
				ARCHIVES = 1;
			} else if (strstr(strrchr(URL->filename, '.'), "zip"))
			{	strcpy(COMMAND, "unzip -t ");
				ARCHIVES = 1;
			} else if (strstr(strrchr(URL->filename, '.'), "arj"))
			{	strcpy(COMMAND, "unarj t ");
				ARCHIVES = 1;
			} else if (strstr(strrchr(URL->filename, '.'), "ace"))
			{	strcpy(COMMAND, "unace -t ");
				ARCHIVES = 1;
			} else if (strstr(strrchr(URL->filename, '.'), "rar"))
			{	strcpy(COMMAND, "unrar -t ");
				ARCHIVES = 1;
			} else if (strstr(strrchr(URL->filename, '.'), "bz2"))
			{	strcpy(COMMAND, "bzip2 -t ");
				ARCHIVES = 1;
			}

			if (ARCHIVES)
				if(system(strcat(COMMAND, URL->filename)) == 0)
					printf("WOOHOO!! %s is a valid archive!\n\n\n", URL->filename);
				else
					printf("OH NO!!! %s HAS ERRORS!!!\n\n\n", URL->filename);
		}
	}
}


void Switches (int i, char *argv[])
/*************************************
**	void Switches (int, char*)
**
** Post: Parses strings in argv and set global variables accordingly.
*************************************/
{	if (argv[i][0] == '-')
	{	switch (argv[i][1])
		{	case 'N':
			case 'n':
				argv[i] = GGR;
				break;
			case 'O':
			case 'o':
				if (argv[i][2] >= '0' || argv[i][2] <= '9')
					OUTPUT_LEVEL = argv[i][2] - '0';
				break;
			case 'R':
			case 'r':
				if (argv[i][2] == '\0')
					REFERRER = 0;
				else
					strcpy(REF, argv[i] + 2);
				break;
			case 'S':
			case 's':
				STDOUT = 1;
				break;
			case 'T':
			case 't':
				TESTFILES = 1;
				break;
			case 'V':
			case 'v':
				printf("%s  [Get and Resume Elite EDition] - By "
				 "Anoakie Turner\r\n", CURRENT_VERSION);
				exit(EXIT_SUCCESS);
				break;
			case 'F':
			case 'f':
				CHECKIN = 0;
				break;
			case 'G':
			case 'g':
				IMPORT_GR = 1;
				GR_FILE = argv[i] + 2;
				break;
			case 'I':
			case 'i':
				STDIN = 1;
				break;
			case 'D':
			case 'd':
				if (argv[i][2] != '\0')
					FORK = atoi(argv[i] + 2);
				else
					FORK = 1;
				break;
			case 'C':
			case 'c':
				argv[i] += 2;
				printf("\n\n<-- %s\n", argv[i]);
				unescape_url(argv[i]);
				printf("--> %s!\n\n", argv[i]);
				break;
			case 'B':
			case 'b':
				BACKGROUND = 1;
				break;
			default:
				printf("\n\n\t\t*** Unknown switch [%s]! ***\n\n", argv[i]);
			case '?':
				printf("%s  [Get and Resume Elite EDition] - By "
				"Anoakie Turner\r\n", CURRENT_VERSION);
				printf("\n\nUSEAGE: greed URL(1) URL(2) ... URL(N)\n"
				"i.e.: greed %s ftp://some.where/out/there.tar.gz\n\n", GGR);
printf("Options:\n"
"-b\tSends greed into background downloading mode.\n"
"-c<url>\tchanges the %%##'s in a URL to characters.\n"
"-d\tTells greed to download 2 files at once.\n"
"-d<#>\tTells greed to download # files at once.\n"
"-f\tDownloads the URLs specified in greed.in.\n"
"-g<FL>\tImports a windows GetRight file (*.GRX).\n"
"\t\ti.e. greed -gMyDownloads.grx\n"
"-i\tReads URLs from standard input.\n"
"\t\ti.e. greed -I < download.lst\n"
"-n\tDownloads the newest version of GREED!\n"
"-o<#>\tSets the level of output (default 5, no output = 0).\n"
"-r\tSets the referrer OFF.\n"
"-r<URL>\tSets the referrer to URL.\n"
"-s\tPrints the file that is being downloaded to the screen instead of a file.\n"
"-t\tTests archives after download to see if they are valid.\n"
"...\tUsage: greed URL...FILE, where URL is the url and FILE is the file retrieved.\n"
"\t\tNOTE: This only works if you have the correct programs to test with!.\n"
"\n\nALL OTHER OPTIONS (including proxy support) CAN BE FOUND IN ~/.greedrc\n");
				exit(0);
				break;
		}
	}
}
